const St = imports.gi.St;
const Main = imports.ui.main;
const Tweener = imports.ui.tweener;
const OverviewControls = imports.ui.overviewControls;
const WorkspacesView = imports.ui.workspacesView;
const Lang = imports.lang;
const GMenu = imports.gi.GMenu;
const GObject = imports.gi.GObject;
const Clutter = imports.gi.Clutter;
const Shell = imports.gi.Shell;
const Gtk = imports.gi.Gtk;
const Pango = imports.gi.Pango;
const ExtensionUtils = imports.misc.extensionUtils;
const Gio = imports.gi.Gio;
const Favorites = imports.ui.appFavorites;
const SCHEMA = 'org.gnome.shell.extensions.activityAppLauncher';
const Gettext = imports.gettext;
const GLib = imports.gi.GLib;

var _ = Gettext.domain("activityAppLauncher").gettext;
Gettext.bindtextdomain("activityAppLauncher", ExtensionUtils.getCurrentExtension().path + "/locale");

// Version 11

var init = function() {
    return new ActivityAppLauncher();
}

class ActivityAppLauncher {

    constructor() {
        this.isEnabled = false;
        this._startupPreparedId = 0;
        this._applicationsButtons = null;
        this.appsInnerContainer = null;
        this.selected = null;
        this.currentActor = null;
        this._appSys = Shell.AppSystem.get_default();
        this._settings = this._get_schema(SCHEMA);
        this.visibility = "atexttoensurethateverythingworks";
    }

    _onChangedSetting() {
        this.icon_size = this._settings.get_int("icon-size");
        this.show_favorites = this._settings.get_boolean("show-favorites");
        this.show_frequent = this._settings.get_boolean("show-frequent");
        this.icon_width = Math.floor(this.icon_size * 2);
        this.icon_height = Math.floor(this.icon_size * 1.8);
    }

    _get_schema (schema) {
        let extension = ExtensionUtils.getCurrentExtension();

        const GioSSS = Gio.SettingsSchemaSource;

        // check if this extension was built with "make zip-file", and thus
        // has the schema files in a subfolder
        // otherwise assume that extension has been installed in the
        // same prefix as gnome-shell (and therefore schemas are available
        // in the standard folders)
        let schemaDir = extension.dir.get_child('schemas');
        let schemaSource;
        if (schemaDir.query_exists(null))
            schemaSource = GioSSS.new_from_directory(schemaDir.get_path(),
                                                     GioSSS.get_default(),
                                                     false);
        else
            schemaSource = GioSSS.get_default();

        let schemaObj = schemaSource.lookup(schema, true);
        if (!schemaObj)
            throw new Error('Schema ' + schema + ' could not be found for extension '
                            + extension.metadata.uuid + '. Please check your installation.');

        return new Gio.Settings({ settings_schema: schemaObj });
    }

    enable() {
        // Wait until startup completed
        if (Main.layoutManager._startingUp) {
            this._startupPreparedId = Main.layoutManager.connect('startup-complete', () => {
                this._doEnable(true);
            });
        } else {
            this._doEnable(false);
        }
    }

    _doEnable(removeStartup) {
        if (removeStartup) {
            Main.layoutManager.disconnect(this._startupPreparedId);
            this._startupPreparedId = 0;
        }
        this._controls = Main.overview._overview._controls;
        this._settingsId = this._settings.connect('changed', () => {this._onChangedSetting();});
        this._pageChangedId = this._controls.viewSelector.connect('page-changed', () => {this._onPageChanged();});
        this._onChangedSetting();
        this.isEnabled = true;
        this.fill_elements(true);
        this._favorites = Favorites.getAppFavorites();
        this._usage = Shell.AppUsage.get_default();
        this.showingId = Main.overview.connect('showing', () => {this._show();});
        this.hidingId = Main.overview.connect('hiding', () => {this._hide();});
    }

    disable() {
        if (this.isEnabled) {
            this.fill_elements(false);
            Main.overview.disconnect(this.showingId);
            Main.overview.disconnect(this.hidingId);
            this._settings.disconnect(this._settingsId);
            this._controls.viewSelector.disconnect(this._pageChangedId);
        }
        this.isEnabled = false;
    }

    _onPageChanged() {
        if ((this._controls.viewSelector._activePage == this.appsLaunchPage) ||
            (this._controls.viewSelector._activePage == this._controls.viewSelector._workspacesPage)) {
                this.appsContainer.show_all();
            } else {
                this.appsContainer.hide();
            }
    }

    _hide() {
        if (this.currentActor !== null) {
            this.appsLaunchContainer.remove_actor(this.currentActor);
            this.currentActor = null;
        }
        this.appsLaunchContainer.hide();
    }

    _show() {
        this.visibility = "atexttoensurethateverythingworks";
        this.selected = null;
        this.setVisibility();
        this.appsLaunchContainer.hide();
        this._fillCategories();
    }

    // Modifies the Activities view to add our stage with the application categories
    fill_elements(desired_mode) {

        if (desired_mode) {
            this.appsContainer = new St.Bin({x_expand: false, y_expand: true, y_fill:true, x_fill: true});
            this.appsLaunchContainer = new St.Bin({y_fill:true, x_fill: true});
            this._controls.viewSelector.show_all();
            this._controls._thumbnailsSlider.show_all();
        } else {
            this._controls._group.remove_actor(this.appsContainer);
            this._controls.viewSelector.remove_actor(this.appsLaunchPage);
            this.appsLaunchPage = null;
            this.appsContainer = null;
            this.appsLaunchContainer = null;
            this.appsInnerContainer = null;
            this.currentActor = null;
        }

        this._controls._group.remove_actor(this._controls.viewSelector);
        this._controls._group.remove_actor(this._controls._thumbnailsSlider);

        if (desired_mode) {
            this._controls._group.add_child(this.appsContainer);
            this.appsLaunchPage = this._controls.viewSelector._addPage(this.appsLaunchContainer,"", '');
        }
        this._controls._group.add_child(this._controls.viewSelector);
        this._controls._group.add_actor(this._controls._thumbnailsSlider);
    }

    _fillCategories() {
        this.selected = null;
        if (this.appsInnerContainer !== null) {
            this.appsContainer.remove_actor(this.appsInnerContainer);
        }
        this.appsInnerContainer = new St.BoxLayout({ vertical: true });
        this.appsInnerContainer.x_align = Clutter.ActorAlign.FILL;
        this.appsInnerContainer.y_align = Clutter.ActorAlign.START;
        this.appsInnerContainer.x_expand = false;
        this.appsInnerContainer.y_expand = false;
        this.appsContainer.add_actor(this.appsInnerContainer);

        this.appsInnerContainer.buttons = [];
        this.appsInnerContainer.appList=[];
        this.appsInnerContainer.appClass=[];

        let tree = new GMenu.Tree({ menu_basename: 'applications.menu' });
        tree.load_sync();
        let root = tree.get_root_directory();
        let categoryMenuItem = new AALCathegory_Menu_Item(this,1,_("Windows"), null);
        this.appsInnerContainer.add_child(categoryMenuItem);
        this.appsInnerContainer.buttons.push(categoryMenuItem);

        if (this.show_favorites) {
            let favoritesMenuItem = new AALCathegory_Menu_Item(this,2,_("Favorites"), null);
            this.appsInnerContainer.add_actor(favoritesMenuItem);
            this.appsInnerContainer.buttons.push(favoritesMenuItem);
        }

        if (this.show_frequent) {
            let mostUsedMenuItem = new AALCathegory_Menu_Item(this,3,_("Frequent"), null);
            this.appsInnerContainer.add_actor(mostUsedMenuItem);
            this.appsInnerContainer.buttons.push(mostUsedMenuItem);
        }

        let iter = root.iter();
        let nextType;
        while ((nextType = iter.next()) != GMenu.TreeItemType.INVALID) {
            if (nextType == GMenu.TreeItemType.DIRECTORY) {
                let dir = iter.get_directory();
                if (!dir.get_is_nodisplay()) {
                    let childrens = this._fillCategories2(dir,[]);
                if (childrens.length != 0) {
                    childrens.sort(this._sortApps);
                    let item = { dirItem: dir, dirChilds: childrens };
                    this.appsInnerContainer.appClass.push(item);
                }
            }
         }
        }
        this.appsInnerContainer.appList.sort(this._sortApps);
        for (var i = 0; i < this.appsInnerContainer.appClass.length; i++) {
            let categoryMenuItem = new AALCathegory_Menu_Item(this,0,this.appsInnerContainer.appClass[i].dirItem.get_name(), this.appsInnerContainer.appClass[i].dirChilds);
            this.appsInnerContainer.add_actor(categoryMenuItem);
            this.appsInnerContainer.buttons.push(categoryMenuItem);
        }
    }

    _sortApps(param1, param2) {
        if (!param1) {
            return 1;
        }
        if (!param2) {
            return -1;
        }
        if (param1.get_name().toUpperCase()<param2.get_name().toUpperCase()) {
            return -1;
        } else {
            return 1;
        }
    }

    _fillCategories2(dir,childrens) {
        let iter = dir.iter();
        let nextType;

        while ((nextType = iter.next()) != GMenu.TreeItemType.INVALID) {
            if (nextType == GMenu.TreeItemType.ENTRY) {
                let entry = iter.get_entry();
                if (!entry.get_app_info().get_nodisplay()) {
                    let app = this._appSys.lookup_app(entry.get_desktop_file_id());
                    childrens.push(app);
                    this.appsInnerContainer.appList.push(app);
                }
            } else if (nextType == GMenu.TreeItemType.DIRECTORY) {
                childrens = this._fillCategories2(iter.get_directory(), childrens);
            }
        }
        return childrens;
    }

    _allocationChanged(button) {
        let [sizex, sizey] = this.currentActor.iconsContainer.get_size();

        if (this.last_iconx >= sizex) {
            return;
        }
        this.last_iconx = sizex;
        var iconx = Math.floor(sizex / this.icon_width);
        this.currentActor.iconsContainer.remove_all_children();
        var position = 0;
        var currentContainer = null;
        var launcherList = null;
        switch (button.launcherType) {
            case 0:
                launcherList = button.launchers;
                break;
            case 2:
                launcherList = this._favorites.getFavorites();
                break;
            case 3:
                launcherList = this._usage.get_most_used();
                break;
            default:
                launcherList = null;
                break;
        }
        if (launcherList !== null) {
            for(let i = 0;i < launcherList.length; i++) {
                var element = launcherList[i];
                if (position == 0) {
                    currentContainer = new St.BoxLayout({vertical: false, x_expand: true});
                    currentContainer.x_expand = true;
                    currentContainer.y_expand = false;
                    this.currentActor.iconsContainer.add_child(currentContainer);
                }
                var tmpContainer = new St.BoxLayout({vertical: true, reactive: true, style_class:'activityAppLauncherElement', width: this.icon_width, height: this.icon_height});
                tmpContainer.icon = element.create_icon_texture(this.icon_size);
                tmpContainer.text = new St.Label({text: element.get_name(), style_class: 'list-search-result-title activityAppLauncherText'});
                tmpContainer.text.clutter_text.line_wrap_mode = Pango.WrapMode.WORD;
                tmpContainer.text.clutter_text.line_wrap = true;
                tmpContainer.set_x_align(Clutter.ActorAlign.CENTER);
                tmpContainer.set_y_align(Clutter.ActorAlign.START);
                tmpContainer.set_x_expand(false);
                tmpContainer.set_y_expand(false);
                tmpContainer.add_child(tmpContainer.icon);
                tmpContainer.text.set_x_align(Clutter.ActorAlign.CENTER);
                tmpContainer.text.set_y_align(Clutter.ActorAlign.START);
                tmpContainer.text.set_x_expand(true);
                tmpContainer.text.set_y_expand(true);
                tmpContainer.add_child(tmpContainer.text);
                currentContainer.add_child(tmpContainer);

                tmpContainer._app = element;
                tmpContainer._customEventId = tmpContainer.connect('button-release-event', Lang.bind(this,
                    function(actor, event) {
                        actor._app.open_new_window(-1);
                        Main.overview.hide();
                    })
                );
                tmpContainer._customEnterId = tmpContainer.connect('enter-event', Lang.bind(this,
                    function (actor, event) {
                        actor.set_style_pseudo_class("selected");
                    }));

                tmpContainer._customLeaveId = tmpContainer.connect('leave-event', Lang.bind(this,
                    function (actor, event) {
                        actor.set_style_pseudo_class("");
                    }));
                position++;
                if (position == iconx) {
                    position = 0;
                }
            }
        }
        this.appsLaunchContainer.show_all();
    }

    clickedCathegory(button) {
        for(var i = 0; i < this.appsInnerContainer.buttons.length; i++) {
            var tmpbutton = this.appsInnerContainer.buttons[i];
            if (button == tmpbutton) {
                tmpbutton.checked = true;
            } else {
                tmpbutton.checked = false;
            }
        }
        if (button.launcherType == 1) {
            this.selected = null; // for Windows list, the selected button must be null
        } else {
            this.selected = button.cat;
        }
        this.setVisibility();
        if (this.currentActor !== null) {
            this.appsLaunchContainer.remove_actor(this.currentActor);
            this.currentActor = null;
        }
        if (this.selected !== null) {
            this.currentActor = new St.ScrollView({hscrollbar_policy: Gtk.PolicyType.NEVER, x_fill: true, y_fill: true});
            this.currentActor.x_expand = true;
            this.currentActor.y_expand = true;
            this.appsLaunchContainer.add_actor(this.currentActor);
            this.currentActor.iconsContainer = new St.BoxLayout({ vertical: true, x_expand: true});
            this.last_iconx = 0;
            this.currentActor.iconsContainer.connect_after("allocation-changed", () => {
                this._allocationChanged(button);
            });
            this.currentActor.iconsContainer.x_fill = true;
            this.currentActor.iconsContainer.y_fill = true;
            this.currentActor.add_actor(this.currentActor.iconsContainer);
            this.appsLaunchContainer.show_all();
        }
    }

    setVisibility() {
        if (this.visibility != this.selected) {
            this.visibility = this.selected;
            var viewSelector = Main.overview._overview._controls.viewSelector;
            if (this.selected === null) {
                viewSelector._showPage(viewSelector._workspacesPage);
            } else {
                viewSelector._showPage(this.appsLaunchPage);
            }
        }
    }
}

const AALCathegory_Menu_Item = GObject.registerClass({
    GTypeName: 'AALCathegory_Menu_Item',
}, class AALCathegory_Menu_Item extends St.Button {

    _init(topClass, type, cathegory, launchers) {
        this.topClass = topClass;
        this.cat = cathegory;
        this.launchers = launchers;
        this.launcherType = type;
        super._init({label: GLib.markup_escape_text(cathegory,-1), style_class: "world-clocks-button button activityAppLauncherButton", toggle_mode: true, can_focus: true, track_hover: true});
        this.connect("clicked",Lang.bind(this,function() {
            this.topClass.clickedCathegory(this);
        }));
        this.show_all();
    }
});
